#include "sa.h"
#include "random.h"
#include "fold.h"
#include <string.h>
#include <stdio.h>
#include "treedist.h"
#include "dist_vars.h"
#include "RNAstruct.h"
#include "fold_vars.h"
/*#include <Brafman.h>*/
#include <stdlib.h>
#include <math.h>

char letters [4] = {'A','U','G','C'};
char  perturbate_table [4][3] = { {'U','G','C'} , {'A','G','C'} , {'A','U','C'} , {'A','U','G'} };

char * target_fold = NULL;
char * target_motif = NULL;
int target_motif_location = -1;
int target_motif_strlen = -1;
char * target_nucleotide = NULL;
int target_nuc_location = -1;
Tree * target_tree = NULL;
double target_dG = (1.0);
double target_mr = (-1.0);

int MAX_TRY = 4;
int tie_ends = 0;
int allow_slip = 0;
int variant_length = 0;

char * random_start (int len)
{
   int i,j;
   char * str;
   str = (char *) malloc(sizeof(char)*(len+1));
   for(i=0 ; i < len ; i++)
   {
      j = (int)(Random()*4.0);
      str[i] = letters[j];
   }
   
   if (target_nucleotide)
   {
      for (i = target_nuc_location; i < (strlen(target_nucleotide) + target_nuc_location); i++)
         str[i] = target_nucleotide[i - target_nuc_location];
   }

   str[len]=0;
   return str;
}


void perturbate (char * str, int len)
{
   int i , j , k;
   i = (int)(Random()*len);
   
   if (target_nucleotide)
   {
	while ((i >= target_nuc_location - allow_slip)&&(i < target_nuc_location + strlen(target_nucleotide) + allow_slip))
   		i = (int)(Random()*len);
   }

   switch (str[i])
   {
   case 'A':
   	j = 0;
   	break;
   case 'U':
   	j = 1;
   	break;
   case 'G':
   	j = 2;
   	break;
   case 'C':
   	j = 3;
   	break;
   }

   k = (int)(Random()*3);

   str[i] = perturbate_table[j][k];   
}

double mutational_robustness(const char * seq)
{
   int i,j,acum,len;
   char let[] = {'A','U','C','G'};
   char org;
   char * new_seq , * structure , * new_structure , * shapiro , * xstruc;
   char old_bt;
   Tree * t1 , * t2;
   //old_bt = backtrack_type;
   //backtrack_type = 'F';
   len = strlen(seq);
   acum = 0;
   new_seq = (char *)malloc(sizeof(char)*(len+1)*10);
   structure = (char *)malloc(sizeof(char)*(len+1)*10);
   new_structure = (char *)malloc(sizeof(char)*(len+1)*10);
   strcpy(new_seq,seq);   
   new_structure[0] = 0;
   fold(seq,structure);

   shapiro = b2C(structure);
   xstruc = unweight(shapiro);
   t1 = make_tree(xstruc);
   free(xstruc);
   free(shapiro);
   
   for (i=0 ; i < len ; i++)
   {
      for(j=0;j<4;j++)
      {
         if(let[j]!=seq[i])
         {
            new_seq[i]=let[j];
            //printf("seq = |%s| , new_seq = |%s|\n",seq,new_seq);
            fold(new_seq,new_structure);
            shapiro = b2C(new_structure);
            xstruc = unweight(shapiro);
            t2 = make_tree(xstruc);
            free(xstruc);
            free(shapiro);
            //printf("%s\n%s\n%d\n",structure,new_structure,hamming(structure,new_structure));
            //if(hamming(structure,new_structure))
            /*if(tree_edit_distance(t1,t2))
               acum++;*/
            acum += bp_distance(structure,new_structure);
            free_tree(t2);
         }
      }
      new_seq[i] = seq[i];
   }
   //backtrack_type = old_bt;
   free_tree(t1);
   //free_tree(t2);
   free(new_seq);
   free(new_structure);
   free(structure);
   //printf("mr = %f\n",1.0-((double)acum)/((double)len*len*3));
   return 1.0-((double)acum)/((double)len*len*3);
}

int does_motif_exist(const char * structure)
{
   char * current_weighted_shapiro;
   char * p_current;
   int motif_found = 0;

   if (target_motif_strlen == -1)
   {
      fprintf(stderr,"No target location\n");
      exit(-5);
   }

   current_weighted_shapiro = b2Shapiro(structure);
   p_current = current_weighted_shapiro;

   // printf("Structure = %s, weighted Shapiro = %s,", structure, current_weighted_shapiro);

   while ((motif_found == 0) && ((p_current = strstr(p_current, target_motif)) != NULL))
   {
    	if (abs(strlen(p_current) - target_motif_strlen) < 3) 
	    	motif_found = 1;
        ++p_current;
   }

   free(current_weighted_shapiro);

   // printf("motif_found = %d\n", motif_found);
   return motif_found;
}

double sa_scoure(char * str , int len)
{
   int i , j=0  ;
   char * structure;
   double dG,accum,shapiro_dist,mr;
   Tree * tree;
   char *xstruc;
   char * shapiro;
   int base_pair_dist;
   int motif_found = 0;

   structure = (char *) malloc(sizeof(char)*(len+1));
   
   if (target_fold == NULL)
   {
      fprintf(stderr,"No target fold\n");
      exit(-4);
   }

   dG = fold(str,structure);

   mr = mutational_robustness(str);
   mr = (target_mr == -1000.0) ? 0.0 : fabs(mr - target_mr)*100.0;
   dG = (target_dG == -1000.0) ? 0.0 : fabs(dG - target_dG);
   
   shapiro = b2C(structure);
   xstruc = unweight(shapiro);
   //printf("Shapiro = %s, Shapiro unweighted = %s, ", shapiro, xstruc);
   tree = make_tree(xstruc);
   free(xstruc);
   free(shapiro);
   shapiro_dist = tree_edit_distance(target_tree, tree);
   if(tie_ends && (structure[0] != '(' || structure[len-1] != ')')) 
	shapiro_dist++;
   free_tree(tree);
  
   base_pair_dist = bp_distance(structure,target_fold);

   //printf("mr = %f, dG = %f, shapiro_dist = %f, base_pair_dist = %d\n", mr, dG, shapiro_dist, base_pair_dist);
   accum = ((double)mr)+dG+shapiro_dist*100+((double)base_pair_dist)*0.01;
   if(target_motif)
   {
	// since it is a minimization function, if a motif is found the accumulated sum is lower
        motif_found = does_motif_exist(structure);
	accum += (1 - motif_found)*1000;
   }   
   free(structure);
   //printf("accum = %f\n", accum);
   return accum;
}

double sa_scoure_print (char * str , int len)
{
   int i , j=0  ;
   char * structure;
   double dG,accum,shapiro_dist,mr;
   Tree * tree;
   char *xstruc;
   char * shapiro;

   structure = (char *) malloc(sizeof(char)*(len+1));
   
   if (target_fold == NULL)
   {
      fprintf(stderr,"No target fold\n");
      exit(-4);
   }

   dG = fold(str,structure);

   mr = mutational_robustness(str);
   //printf("raw mr = %f\n",mr);
   
   shapiro = b2C(structure);
   xstruc = unweight(shapiro);
   //printf("xstruc = %s\n",xstruc);
   tree = make_tree(xstruc);
   free(xstruc);
   printf("Output shape in Shapiro representation = %s\n",shapiro); 
   free(shapiro);
   shapiro_dist = tree_edit_distance(target_tree, tree);
   free_tree(tree); 
   free(structure);

   printf("Output mutational robustness = %f\nOutput minimum free energy (Kcal/mol) = %f\n",mr,dG);
   mr = (target_mr == -1000.0) ? 0.0 : fabs(mr - target_mr)*100.0;
   dG = (target_dG == -1000.0) ? 0.0 : fabs(dG - target_dG);
   accum = ((double)mr)+dG+shapiro_dist*100;
   return accum;
}

void validate_input()
{
   char * weighted_shapiro;
   int i;

   weighted_shapiro = b2Shapiro(target_fold);
   if (target_motif && strstr(weighted_shapiro, target_motif) == NULL)
   {
      fprintf(stderr,"No target motif in target fold\n");
      exit(-5);
   }
   free(weighted_shapiro);

   if (target_motif && target_motif_location == -1)
   {
      fprintf(stderr,"No target location\n");
      exit(-6);
   }

   if (target_nucleotide && target_nuc_location == -1)
   {
      fprintf(stderr,"No target nucleotide location specified\n");
      exit(-7);
   }

   if (target_nucleotide && strlen(target_nucleotide) == 0)
   {
      fprintf(stderr,"Target nucleotide with zero length\n");
      exit(-10);
   }

   if (target_nucleotide && (strlen(target_nucleotide) + target_nuc_location > strlen(target_fold)))
   {
      fprintf(stderr,"Target nucleotide exceeds fold length\n");
      exit(-8);
   }

   if (target_nucleotide)
   {
	for (i = 0; i < strlen(target_nucleotide); i++)
	{
	   switch (target_nucleotide[i])
	   {
	   case 'A':
	   case 'U':
	   case 'G':
	   case 'C':
	      break;
	   default:
	      fprintf(stderr,"Illegal nucleotide constraint\n");
	      exit(-9);
	      break;
	   }
	}
   }
}

char * simulated_anilling (int iteretion_number , int len)
{
   char * seq = NULL, * seq_copy = NULL, * best = NULL, * start = NULL;
   int i,j;
   int variant = 0;
   double current_scoure , new_scoure ,dG , best_scoure;
   char *xstruc , * shapiro;
   char * weighted_shapiro;
   int nucleotide_found = 0;
   char * p_current = NULL;

   if (target_fold == NULL)
   {
      fprintf(stderr,"No target fold\n");
      exit(-4);
   }

   noLonelyPairs = 1;
   initialize_fold(len);
   initRandom();

   shapiro = b2C(target_fold);
   /* Debug print
   printf("Target shape in Shapiro representation = %s\n",shapiro);
   printf("Target mutational robustness = %f\n",target_mr);
   printf("Target minimum free energy (Kcal/mol)= %f\n",target_dG);
   */
   xstruc = unweight(shapiro);
   target_tree = make_tree(xstruc);
   free(xstruc);
   free(shapiro);

   validate_input();

   weighted_shapiro = b2Shapiro(target_fold);
   if (target_motif)
   {
      target_motif_strlen = strlen(weighted_shapiro) - target_motif_location;
   }

   free(weighted_shapiro);

   seq_copy = (char *)malloc(sizeof(char)*(len+1));
   best = (char *)malloc(sizeof(char)*(len+1));

   for(j = MAX_TRY + 1 ; j > 0 ; j--)
   {
        if (start) // Allocated in previous iteration and never freed
	    free(start);
   	start = random_start(len); 
   	inverse_fold(start, target_fold);   
   
   	seq = start;

	p_current = seq;
        nucleotide_found = 0;
   	while (target_nucleotide && (nucleotide_found == 0) && (p_current = strstr(p_current, target_nucleotide)))
   	{
	/*
	   printf("len = %d\n", len);
	   printf("strlen(p_current) = %zd\n", strlen(p_current));
	   printf("target_nuc_location = %d\n", target_nuc_location);
	   printf("abs(len - strlen(p_current) - target_nuc_location) = %d\n", abs(len - strlen(p_current) - target_nuc_location));
	*/    	
	   if (abs(len - strlen(p_current) - target_nuc_location) < (allow_slip + 1)) 
	    	nucleotide_found = 1;
	// printf("p_current = %s\n", p_current);
           ++p_current;
	}

	if (target_nucleotide && nucleotide_found == 0)
		continue; //continue to next since nucleotide is not in the initial sequence 
	
   	current_scoure = sa_scoure(seq,len);

        if (current_scoure < 1000)
	{
	   //printf("%s\n", seq);
	   if (variant_length)
		variant = (int)(Random()*2.0);
	   switch (variant)
	   {
	   case 0:
		strcpy(seq_copy,seq);
		break;
	   case 1:
		strncpy(seq_copy, seq + 1, strlen(seq) - 2);
		seq_copy[strlen(seq)-2] = '\0';
		strcpy(seq,seq_copy);
		current_scoure = sa_scoure(seq,strlen(seq));
		break;
	   }
	   //printf("%s\n", seq_copy);
	   
	   if (current_scoure < 1000)
		break;
	}
   }
   //printf("%f\n",current_scoure);
   best_scoure = current_scoure;
   strcpy(best,seq);

   if (target_motif && best_scoure >= 1000)
   {
      fprintf(stderr,"No target motif in initial sequence\n");
      exit(-10);
   }

   if (target_nucleotide && nucleotide_found == 0) 
   {
      fprintf(stderr,"No nucleotide in initial sequence\n");
      exit(-11);      
   }

   /* Debug print
   sa_scoure_print(seq_copy,len);
   */

   if ((variant_length == 0) && (strlen(seq_copy) != len))
   {
      fprintf(stderr,"Data corruption\n");
      exit(-12);      
   }
   
   for (i = 0 ; i < iteretion_number ; i++)
   {
      for(j = MAX_TRY ; j > 0 ; j--)
      {
         perturbate(seq_copy,strlen(seq_copy));
         new_scoure = sa_scoure(seq_copy,strlen(seq_copy));
        /* Debug print
         sa_scoure_print(seq_copy,strlen(seq_copy));
	*/
         if (new_scoure < current_scoure)
            break;
         if(Random() < (((double)2.0)/((double)i+1))/((double)MAX_TRY) && new_scoure < 100)
	 {
	    //printf("random\n");	
            break;
	 }
      }
      if (!j)
         strcpy(seq_copy,seq);
      else
      {
         current_scoure = new_scoure;
         if(current_scoure == 0.0)
            break;
         if(current_scoure < best_scoure)
         {
            best_scoure = current_scoure;
            strcpy(best,seq_copy);
            /* Debug print
            sa_scoure_print(seq_copy,strlen(seq_copy));
	    */
         }               
      }

      strcpy(seq,seq_copy);
      //if (i % 100 == 0)
         //printf("iteration %d done soure = %f j = %d\n",i,current_scoure,j);
   }

   //printf("%f\n", best_scoure);
   sa_scoure_print(best,strlen(best));
   /* Debug print
   printf("END best = %f\n",best_scoure);
   printf("Best seq = %s\n",best);
   */
   free(seq);
   //Brafman;
   free(seq_copy);
   //Brafman;
   free_tree(target_tree);
   //Brafman;
   return best;
}

